/**
 * @file Contains helper functions for Bootstrap Accordion.
 */

import type {
  DocumentSelection,
  Element,
  ViewDocumentSelection,
  ViewElement,
  Writer
} from 'ckeditor5/src/engine';
import type { AccordionCollapseModelAttribute, AccordionButtonModelAttribute } from './bootstrapaccordionconfig'
import { collapseShowDefault, buttonCollapsedDefault } from './bootstrapaccordionconfig';

/**
 * Checks if a provided element is an accordion widget.
 *
 * @param element
 *   The element to check.
 * @returns
 *   Whether the element is an accordion widget.
 */
export function isAccordionWidget(element: ViewElement): boolean {
  return element.hasClass('ckeditor5-bootstrap-accordion__widget');
}

/**
 * Gets the selected accordion widget.
 *
 * @param selection
 *   The view document selection.
 * @returns
 *   The selected accordion widget, or null if there is none.
 */
export function getSelectedAccordionWidget(selection: ViewDocumentSelection): ViewElement | null {
  const accordion = selection.focus?.getAncestors().reverse()
    .find(node => node.is('element') && isAccordionWidget(node));
  return accordion?.is('element') ? accordion : null;
}

/**
 * Gets the selected accordion item model element.
 *
 * @param selection
 *   The model document selection.
 * @returns
 *   The selected accordion item element, or null if there is none.
 */
export function getSelectedAccordionItemModelElement(selection: DocumentSelection): Element | null | undefined {
  return selection.getFirstPosition()?.findAncestor('bootstrapAccordionItem');
}

/**
 * Gets the selected accordion model element.
 *
 * @param selection
 *   The model document selection.
 * @returns
 *   The selected accordion element, or null if there is none.
 */
export function getSelectedAccordionModelElement(selection: DocumentSelection): Element | null | undefined {
  return selection.getFirstPosition()?.findAncestor('bootstrapAccordion');
}

/**
 * Creates a `<accordionItem>` model element with the necessary child elements.
 *
 * @param writer
 *   The model writer.
 * @returns
 *   The accordion item model element.
 */
export function createAccordionItem(writer: Writer) {
  const accordionItem = writer.createElement('bootstrapAccordionItem');
  const accordionHeader = writer.createElement('bootstrapAccordionHeader');
  const accordionButton = writer.createElement('bootstrapAccordionButton', {
    bootstrapAccordionButtonCollapsed: buttonCollapsedDefault,
  } as { [key in AccordionButtonModelAttribute]: string; });
  const accordionCollapse = writer.createElement('bootstrapAccordionCollapse', {
    bootstrapAccordionCollapseShow: collapseShowDefault,
  } as { [key in AccordionCollapseModelAttribute]: string; });
  const accordionBody = writer.createElement('bootstrapAccordionBody');

  writer.append(accordionHeader, accordionItem);
  writer.append(accordionButton, accordionHeader);
  writer.append(accordionCollapse, accordionItem);
  writer.append(accordionBody, accordionCollapse);

  // The accordionBody text content will automatically be wrapped in a
  // `<p>`.
  writer.appendElement('paragraph', accordionBody);

  return { accordionItem, accordionHeader, accordionButton, accordionCollapse, accordionBody };
}
